"use strict";
/* eslint-disable functional/no-return-void */
/* eslint-disable sonarjs/no-duplicate-string */
/* eslint-disable jest/no-conditional-expect */
/* eslint-disable @typescript-eslint/no-unsafe-call */
/* eslint-disable functional/no-conditional-statements */
/* eslint-disable functional/no-loop-statements */
/* eslint-disable functional/functional-parameters */
/* eslint-disable functional/no-expression-statements */
Object.defineProperty(exports, "__esModule", { value: true });
const _1 = require(".");
describe("ReadonlyURL", () => {
    it("iterates through URL search params using for..of", () => {
        const url = (0, _1.readonlyURL)("http://example.com?foo=a&bar=b");
        if (url === undefined) {
            // eslint-disable-next-line functional/no-throw-statements
            throw new Error("url was undefined");
        }
        for (const [k, v] of url.searchParams) {
            if (k === "foo") {
                expect(v).toEqual("a");
            }
            else {
                expect(k).toEqual("bar");
                expect(v).toEqual("b");
            }
        }
    });
    it("doesn't throw on invalid input", () => {
        const url = (0, _1.readonlyURL)("asdf");
        expect(url).toBeUndefined();
    });
    it("doesn't allow mutation via search params", () => {
        const url = (0, _1.readonlyURL)("http://example.com");
        expect(url).toBeDefined();
        if (url === undefined) {
            // eslint-disable-next-line functional/no-throw-statements
            throw new Error("url was undefined");
        }
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-expect-error
        url.searchParams.append("", "");
    });
});
describe("ReadonlyURLSearchParams", () => {
    it("doesn't throw", () => {
        const x = (0, _1.readonlyURLSearchParams)("");
        expect(x).toBeDefined();
    });
    it("allows iteration using for..of", () => {
        const params = (0, _1.readonlyURLSearchParams)({ a: "b" });
        expect(params[Symbol.iterator]).toBeDefined();
        for (const [k, v] of params) {
            expect(k).toBe("a");
            expect(v).toBe("b");
        }
        /* eslint-disable functional/immutable-data, @typescript-eslint/no-unsafe-assignment, @typescript-eslint/consistent-type-assertions, @typescript-eslint/ban-ts-comment */
        // @ts-expect-error
        params[Symbol.iterator] = null;
        /* eslint-enable functional/immutable-data, @typescript-eslint/no-unsafe-assignment, @typescript-eslint/consistent-type-assertions, @typescript-eslint/ban-ts-comment */
    });
    // eslint-disable-next-line jest/expect-expect
    it("does not expose forEach", () => {
        const params = (0, _1.readonlyURLSearchParams)({ a: "b" });
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-expect-error
        // eslint-disable-next-line @typescript-eslint/no-empty-function
        params.forEach(() => { });
    });
});
describe("ReadonlyDate", () => {
    it("doesn't throw on invalid input", () => {
        const date = (0, _1.readonlyDate)("asdf");
        expect(date.getMilliseconds()).toBeNaN();
    });
    it("doesn't throw when getting current date (now)", () => {
        const date = (0, _1.readonlyDate)((0, _1.readonlyNow)());
        expect(date.getMilliseconds()).toBeGreaterThan(0);
    });
});
describe("ValidReadonlyDate", () => {
    it("returns undefined on invalid input", () => {
        const date = (0, _1.validReadonlyDate)("asdf");
        expect(date).toBeUndefined();
    });
    it("returns a date for valid input", () => {
        // eslint-disable-next-line no-restricted-globals
        const date = (0, _1.validReadonlyDate)(Date.now());
        expect(date).toBeDefined();
    });
});
describe("ReadonlyMap", () => {
    it("allows iteration using for..of", () => {
        const map = (0, _1.readonlyMap)([
            [1, "one"],
            [2, "two"],
            [3, "three"],
        ]);
        for (const [k, v] of map) {
            expect(k).toBeTruthy();
            expect(v).toBeTruthy();
        }
    });
});
describe("ReadonlyWeakMap", () => {
    it("doesn't throw", () => {
        const map = (0, _1.readonlyWeakMap)([
            [{}, "one"],
            [{}, "two"],
            [{}, "three"],
        ]);
        expect(map).toBeDefined();
    });
});
describe("ReadonlySet", () => {
    it("allows iteration using for..of", () => {
        const set = (0, _1.readonlySet)([
            [1, "one"],
            [2, "two"],
            [3, "three"],
        ]);
        for (const [k, v] of set) {
            expect(k).toBeTruthy();
            expect(v).toBeTruthy();
        }
    });
});
describe("ReadonlyWeakSet", () => {
    it("doesn't throw", () => {
        const set = (0, _1.readonlyWeakSet)([
            [{}, "one"],
            [{}, "two"],
            [{}, "three"],
        ]);
        expect(set).toBeDefined();
    });
});
describe("ReadonlyPromise", () => {
    // eslint-disable-next-line functional/prefer-immutable-types
    it("can be awaited", async () => {
        const p = Promise.resolve("a");
        const result = await p;
        expect(result).toStrictEqual("a");
        return undefined;
    });
});
describe("PrincipledArray", () => {
    it("can be mapped", () => {
        const foo = (0, _1.principledArray)(["a"]);
        expect(foo.map((s, _i, _array) => s.toUpperCase())).toStrictEqual(["A"]);
    });
    it("can be flatMapped", () => {
        const foo = (0, _1.principledArray)(["a"]);
        expect(foo.flatMap((s, _i, _array) => (0, _1.principledArray)([s.toUpperCase()]))).toStrictEqual(["A"]);
    });
    it("can be flattened", () => {
        const foo = (0, _1.principledArray)([
            (0, _1.principledArray)(["a"]),
            (0, _1.principledArray)(["b"]),
        ]);
        const flattened = foo.flat();
        expect(flattened).toStrictEqual(["a", "b"]);
    });
    it("can be flattened when nested level is regular array", () => {
        // eslint-disable-next-line functional/prefer-immutable-types
        const foo = (0, _1.principledArray)([["a"], ["b"]]);
        const flattened = foo.flat();
        expect(flattened).toStrictEqual(["a", "b"]);
    });
    it("can be flattened when nested level is mixed", () => {
        // eslint-disable-next-line functional/prefer-immutable-types
        const foo = (0, _1.principledArray)([["a"], (0, _1.principledArray)(["b"]), "c"]);
        const flattened = foo.flat();
        expect(flattened).toStrictEqual(["a", "b", "c"]);
    });
    it("can be filtered with a boolean predicate", () => {
        const foo = (0, _1.principledArray)(["a", "b"]);
        expect(foo.filter((s, _index, _array) => s !== "b")).toStrictEqual(["a"]);
    });
    it("can be filtered with a type guard predicate", () => {
        const foo = (0, _1.principledArray)(["a", "b"]);
        const isLetterA = (s) => s === "a";
        const aArray = foo.filter(isLetterA);
        expect(aArray).toStrictEqual(["a"]);
    });
    it("can use find with a boolean predicate", () => {
        const foo = (0, _1.principledArray)(["a", "b"]);
        expect(foo.find((s, _index, _array) => s !== "b")).toStrictEqual("a");
    });
    it("can use find with a type guard predicate", () => {
        const foo = (0, _1.principledArray)(["a", "b"]);
        const isLetterA = (s) => s === "a";
        const result = foo.find(isLetterA);
        expect(result).toStrictEqual("a");
    });
    it("can use findIndex", () => {
        const foo = (0, _1.principledArray)(["a", "b"]);
        expect(foo.findIndex((s, _index, _array) => s !== "b")).toStrictEqual(0);
    });
    // eslint-disable-next-line jest/expect-expect
    it("does not expose forEach", () => {
        const foo = (0, _1.principledArray)(["a", "b"]);
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-expect-error
        // eslint-disable-next-line @typescript-eslint/no-empty-function
        foo.forEach(() => { });
    });
    it("can be concatenated with a single value", () => {
        const foo = (0, _1.principledArray)(["a"]);
        const bar = foo.concat("b");
        expect(bar).toStrictEqual(["a", "b"]);
    });
    it("can be concatenated with another array", () => {
        const foo = (0, _1.principledArray)(["a"]);
        const bar = foo.concat(["b"]);
        expect(bar).toStrictEqual(["a", "b"]);
    });
    it("can be concatenated with a spread array", () => {
        const foo = (0, _1.principledArray)(["a"]);
        const bar = foo.concat(...["b"]);
        expect(bar).toStrictEqual(["a", "b"]);
    });
    it("can be concatenated with a principled array", () => {
        const foo = (0, _1.principledArray)(["a"]);
        const bar = foo.concat((0, _1.principledArray)(["b"]));
        expect(bar).toStrictEqual(["a", "b"]);
    });
    it("can be sliced", () => {
        const foo = (0, _1.principledArray)(["a"]);
        const bar = foo.slice(0, 1);
        expect(bar).toStrictEqual(["a"]);
    });
    it("can use every", () => {
        const foo = (0, _1.principledArray)(["a"]);
        const bar = foo.every((s, _index, _array) => {
            return s === "a";
        });
        expect(bar).toBe(true);
    });
    it("can use some", () => {
        const foo = (0, _1.principledArray)(["a"]);
        const bar = foo.some((s, _index, _array) => {
            return s === "a";
        });
        expect(bar).toBe(true);
    });
    it("can use reduce", () => {
        const foo = (0, _1.principledArray)(["a", "b"]);
        const bar = foo.reduce((p, c, _index, _array) => `${p}${c}`, "");
        expect(bar).toBe("ab");
    });
    it("can use reduceRight", () => {
        const foo = (0, _1.principledArray)(["a", "b"]);
        const bar = foo.reduceRight((p, c, _index, _array) => `${p}${c}`, "");
        expect(bar).toBe("ba");
    });
    it("cannot use reduce to cause a runtime error with an empty array", () => {
        const foo = (0, _1.principledArray)([]);
        const unsafe = () => 
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-expect-error
        foo.reduce((_p, c, _index, _array) => c);
        expect(unsafe).toThrow();
    });
    it("cannot use reduceRight to cause a runtime error with an empty array", () => {
        const foo = (0, _1.principledArray)([]);
        const unsafe = () => 
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-expect-error
        foo.reduceRight((_p, c, _index, _array) => c);
        expect(unsafe).toThrow();
    });
    it("can use reduce without an initialValue if verifiably nonempty", () => {
        const foo = (0, _1.principledNonEmptyArray)(["a"]);
        const head = foo[0];
        expect(head).toStrictEqual("a");
        expect(foo.reduce((_p, c, _index, _array) => c)).toStrictEqual("a");
    });
    it("can use reduceRight without an initialValue if verifiably nonempty", () => {
        const foo = (0, _1.principledNonEmptyArray)(["a"]);
        const head = foo[0];
        expect(head).toStrictEqual("a");
        expect(foo.reduceRight((_p, c, _index, _array) => c)).toStrictEqual("a");
    });
});
